@@ -42,11 +42,27 @@ _gdbm_new_bucket (GDBM_FILE dbf, hash_bucket *bucket, int bits)
     bucket->h_table[index].hash_value = -1;
 }
 
+/* Return true if the directory entry at DIR_INDEX can be considered
+   valid. This means that DIR_INDEX is in the valid range for addressing
+   the dir array, and the offset stored in dir[DIR_INDEX] points past
+   first two blocks in file. This does not necessarily mean that there's
+   a valid bucket or data block at that offset. All this implies is that
+   it is safe to use the offset for look up in the bucket cache and to
+   attempt to read a block at that offset. */
+int
+gdbm_dir_entry_valid_p (GDBM_FILE dbf, int dir_index)
+{
+  return dir_index >= 0
+         && dir_index < GDBM_DIR_COUNT (dbf)
+         && dbf->dir[dir_index] >= 2*dbf->header->block_size;
+}
+    
 /* Find a bucket for DBF that is pointed to by the bucket directory from
    location DIR_INDEX.   The bucket cache is first checked to see if it
    is already in memory.  If not, a bucket may be tossed to read the new
-   bucket.  In any case, the requested bucket becomes the "current" bucket
-   and dbf->bucket points to the correct bucket. */
+   bucket.  On success, the requested bucket becomes the "current" bucket
+   and dbf->bucket points to the correct bucket. On error, the current
+   bucket remains unchanged. */
 
 int
 _gdbm_get_bucket (GDBM_FILE dbf, int dir_index)
@@ -56,6 +72,13 @@ _gdbm_get_bucket (GDBM_FILE dbf, int dir_index)
   off_t	file_pos;	/* The return address for lseek. */
   int   index;		/* Loop index. */
 
+  if (!gdbm_dir_entry_valid_p (dbf, dir_index))
+    {
+      /* FIXME: negative caching? */
+      GDBM_SET_ERRNO (dbf, GDBM_BAD_DIR_ENTRY, TRUE);
+      return -1;
+    }
+  
   /* Initial set up. */
   dbf->bucket_dir = dir_index;
   bucket_adr = dbf->dir[dir_index];
@@ -69,9 +92,12 @@ _gdbm_get_bucket (GDBM_FILE dbf, int dir_index)
 	}
     }
 
-  /* Is that one is not already current, we must find it. */
+  /* If that one is not already current, we must find it. */
   if (dbf->cache_entry->ca_adr != bucket_adr)
     {
+      size_t lru;
+      hash_bucket *bucket;
+      
       /* Look in the cache. */
       for (index = 0; index < dbf->cache_size; index++)
         {
@@ -84,30 +110,28 @@ _gdbm_get_bucket (GDBM_FILE dbf, int dir_index)
         }
 
       /* It is not in the cache, read it from the disk. */
-      dbf->last_read = (dbf->last_read + 1) % dbf->cache_size;
-      if (dbf->bucket_cache[dbf->last_read].ca_changed)
-	{
-	  if (_gdbm_write_bucket (dbf, &dbf->bucket_cache[dbf->last_read]))
-	    return -1;
-	}
-      dbf->bucket_cache[dbf->last_read].ca_adr = bucket_adr;
-      dbf->bucket = dbf->bucket_cache[dbf->last_read].ca_bucket;
-      dbf->cache_entry = &dbf->bucket_cache[dbf->last_read];
-      dbf->cache_entry->ca_data.elem_loc = -1;
-      dbf->cache_entry->ca_changed = FALSE;
 
-      /* Read the bucket. */
-      file_pos = GDBM_DEBUG_OVERRIDE ("_gdbm_get_bucket:seek-failure",
-				      __lseek (dbf, bucket_adr, SEEK_SET));
+      /* Position the file pointer */
+      file_pos = gdbm_file_seek (dbf, bucket_adr, SEEK_SET);
       if (file_pos != bucket_adr)
 	{
-	  _gdbm_fatal (dbf, _("lseek error"));
 	  GDBM_SET_ERRNO (dbf, GDBM_FILE_SEEK_ERROR, TRUE);
+	  _gdbm_fatal (dbf, _("lseek error"));
 	  return -1;
 	}
       
-      rc = GDBM_DEBUG_OVERRIDE ("_gdbm_get_bucket:read-failure",
-		_gdbm_full_read (dbf, dbf->bucket, dbf->header->bucket_size));
+      /* Flush and drop the last recently used cache entry */
+      lru = (dbf->last_read + 1) % dbf->cache_size;
+      if (dbf->bucket_cache[lru].ca_changed)
+	{
+	  if (_gdbm_write_bucket (dbf, &dbf->bucket_cache[lru]))
+	    return -1;
+	}
+      _gdbm_cache_entry_invalidate (dbf, lru);
+      
+      /* Read the bucket. */
+      rc = _gdbm_full_read (dbf, dbf->bucket_cache[lru].ca_bucket,
+			    dbf->header->bucket_size);
       if (rc)
 	{
 	  GDBM_DEBUG (GDBM_DEBUG_ERR,
@@ -117,8 +141,28 @@ _gdbm_get_bucket (GDBM_FILE dbf, int dir_index)
 	  _gdbm_fatal (dbf, gdbm_db_strerror (dbf));
 	  return -1;
 	}
-    }
+      /* Validate the bucket */
+      bucket = dbf->bucket_cache[lru].ca_bucket;
+      if (!(bucket->count >= 0
+	    && bucket->count <= dbf->header->bucket_elems
+	    && bucket->bucket_bits >= 0
+	    && bucket->bucket_bits <= dbf->header->dir_bits))
+	{
+	  GDBM_SET_ERRNO (dbf, GDBM_BAD_BUCKET, TRUE);
+	  return -1;
+	}
+      /* Validate bucket_avail table */
+      if (gdbm_bucket_avail_table_validate (dbf, bucket))
+	return -1;
 
+      /* Finally, store it in cache */
+      dbf->last_read = lru;
+      dbf->bucket_cache[lru].ca_adr = bucket_adr;
+      dbf->bucket = dbf->bucket_cache[lru].ca_bucket;
+      dbf->cache_entry = &dbf->bucket_cache[lru];
+      dbf->cache_entry->ca_data.elem_loc = -1;
+      dbf->cache_entry->ca_changed = FALSE;
+    }
   return 0;
 }
 
@@ -146,7 +190,7 @@ _gdbm_read_bucket_at (GDBM_FILE dbf, off_t off, hash_bucket *bucket,
     }
 
   /* Read the bucket. */
-  file_pos = __lseek (dbf, off, SEEK_SET);
+  file_pos = gdbm_file_seek (dbf, off, SEEK_SET);
   if (file_pos != off)
     {
       GDBM_SET_ERRNO (dbf, GDBM_FILE_SEEK_ERROR, TRUE);
@@ -194,7 +238,7 @@ _gdbm_split_bucket (GDBM_FILE dbf, int next_insert)
   int          elem_loc;	/* Location in new bucket to put element. */
   bucket_element *old_el;	/* Pointer into the old bucket. */
   int	       select;		/* Used to index bucket during movement. */
-
+  
   /* No directories are yet old. */
   old_count = 0;
 
@@ -259,8 +303,7 @@ _gdbm_split_bucket (GDBM_FILE dbf, int next_insert)
 	  dir_adr  = _gdbm_alloc (dbf, dir_size);
 	  if (dir_adr == 0)
 	    return -1;
-	  new_dir = GDBM_DEBUG_ALLOC ("_gdbm_split_bucket:malloc-failure",
-				      malloc (dir_size));
+	  new_dir = malloc (dir_size);
 	  if (new_dir == NULL)
 	    {
 	      GDBM_SET_ERRNO (dbf, GDBM_MALLOC_ERROR, TRUE);
@@ -376,7 +419,8 @@ _gdbm_split_bucket (GDBM_FILE dbf, int next_insert)
 
   /* Get rid of old directories. */
   for (index = 0; index < old_count; index++)
-    _gdbm_free (dbf, old_adr[index], old_size[index]);
+    if (_gdbm_free (dbf, old_adr[index], old_size[index]))
+      return -1;
 
   return 0;
 }
@@ -391,16 +435,14 @@ _gdbm_write_bucket (GDBM_FILE dbf, cache_elem *ca_entry)
   int rc;
   off_t file_pos;	/* The return value for lseek. */
 
-  file_pos = GDBM_DEBUG_OVERRIDE ("_gdbm_write_bucket:seek-failure",
-				  __lseek (dbf, ca_entry->ca_adr, SEEK_SET));
+  file_pos = gdbm_file_seek (dbf, ca_entry->ca_adr, SEEK_SET);
   if (file_pos != ca_entry->ca_adr)
     {
       GDBM_SET_ERRNO (dbf, GDBM_FILE_SEEK_ERROR, TRUE);
       _gdbm_fatal (dbf, _("lseek error"));
       return -1;
     }
-  rc = GDBM_DEBUG_OVERRIDE ("_gdbm_write_bucket:write-failure",
-        _gdbm_full_write (dbf, ca_entry->ca_bucket, dbf->header->bucket_size));
+  rc = _gdbm_full_write (dbf, ca_entry->ca_bucket, dbf->header->bucket_size);
   if (rc)
     {
       GDBM_DEBUG (GDBM_DEBUG_STORE|GDBM_DEBUG_ERR,
