@@ -43,6 +43,18 @@ int open_mode;                /* Default open mode */
 unsigned input_line;
 
 
+static void
+closedb (void)
+{
+  if (gdbm_file)
+    {
+      gdbm_close (gdbm_file);
+      gdbm_file = NULL;
+      free (file_name);
+      file_name = NULL;
+    }
+}
+
 static int
 opendb (char *dbname)
 {
@@ -78,7 +90,7 @@ opendb (char *dbname)
       
   if (open_mode == GDBM_NEWDB)
     {
-      if (interactive && variable_is_true ("confirm") &&
+      if (interactive () && variable_is_true ("confirm") &&
 	  access (dbname, F_OK) == 0)
 	{
 	  if (!getyn (_("database %s already exists; overwrite"), dbname))
@@ -110,7 +122,7 @@ opendb (char *dbname)
 }
 
 static int
-checkdb ()
+checkdb (void)
 {
   if (!gdbm_file)
     {
@@ -124,6 +136,13 @@ checkdb ()
     }
   return 0;
 }
+
+static int
+checkdb_begin (struct handler_param *param GDBM_ARG_UNUSED,
+	       size_t *exp_count GDBM_ARG_UNUSED)
+{
+  return checkdb ();
+}
 
 size_t
 bucket_print_lines (hash_bucket *bucket)
@@ -195,7 +214,6 @@ _gdbm_avail_list_size (GDBM_FILE dbf, size_t min_size)
   int             size;
   avail_block    *av_stk;
   size_t          lines;
-  int             rc;
   
   lines = 4 + dbf->header->avail.count;
   if (lines > min_size)
@@ -209,25 +227,24 @@ _gdbm_avail_list_size (GDBM_FILE dbf, size_t min_size)
   /* Traverse the stack. */
   while (temp)
     {
-      if (__lseek (dbf, temp, SEEK_SET) != temp)
+      if (gdbm_file_seek (dbf, temp, SEEK_SET) != temp)
 	{
 	  terror ("lseek: %s", strerror (errno));
 	  break;
 	}
       
-      if ((rc = _gdbm_full_read (dbf, av_stk, size)))
+      if (_gdbm_full_read (dbf, av_stk, size))
 	{
-	  if (rc == GDBM_FILE_EOF)
-	    terror ("read: %s", gdbm_strerror (rc));
-	  else
-	    terror ("read: %s (%s)",
-			  gdbm_strerror (rc), strerror (errno));
+	  terror ("read: %s", gdbm_db_strerror (dbf));
 	  break;
 	}
 
-      lines += av_stk->count;
-      if (lines > min_size)
-	break;
+      if (gdbm_avail_block_valid_p (av_stk))
+	{
+	  lines += av_stk->count;
+	  if (lines > min_size)
+	    break;
+	}
       temp = av_stk->next_block;
     }
   free (av_stk);
@@ -235,56 +252,58 @@ _gdbm_avail_list_size (GDBM_FILE dbf, size_t min_size)
   return lines;
 }
 
+static void
+av_table_display (avail_elem *av_table, int count, FILE *fp)
+{
+  int i;
+  
+  for (i = 0; i < count; i++)
+    {
+      fprintf (fp, "  %15d   %10lu \n",
+	       av_table[i].av_size, (unsigned long) av_table[i].av_adr);
+    }
+}
+
 void
 _gdbm_print_avail_list (FILE *fp, GDBM_FILE dbf)
 {
   int             temp;
   int             size;
   avail_block    *av_stk;
-  int             rc;
   
   /* Print the the header avail block.  */
   fprintf (fp, _("\nheader block\nsize  = %d\ncount = %d\n"),
 	   dbf->header->avail.size, dbf->header->avail.count);
-  for (temp = 0; temp < dbf->header->avail.count; temp++)
-    {
-      fprintf (fp, "  %15d   %10lu \n",
-	       dbf->header->avail.av_table[temp].av_size,
-	       (unsigned long) dbf->header->avail.av_table[temp].av_adr);
-    }
+  av_table_display (dbf->header->avail.av_table, dbf->header->avail.count, fp);
 
   /* Initialize the variables for a pass throught the avail stack. */
   temp = dbf->header->avail.next_block;
-  size = (((dbf->header->avail.size * sizeof (avail_elem)) >> 1)
-	  + sizeof (avail_block));
+  size = (dbf->header->avail.size * sizeof (avail_elem))
+	  + sizeof (avail_block);
   av_stk = emalloc (size);
 
   /* Print the stack. */
   while (temp)
     {
-      if (__lseek (dbf, temp, SEEK_SET) != temp)
+      if (gdbm_file_seek (dbf, temp, SEEK_SET) != temp)
 	{
 	  terror ("lseek: %s", strerror (errno));
 	  break;
 	}
       
-      if ((rc = _gdbm_full_read (dbf, av_stk, size)))
+      if (_gdbm_full_read (dbf, av_stk, size))
 	{
-	  if (rc == GDBM_FILE_EOF)
-	    terror ("read: %s", gdbm_strerror (rc));
-	  else
-	    terror ("read: %s (%s)", gdbm_strerror (rc), strerror (errno));
+          terror ("read: %s", gdbm_db_strerror (dbf));
 	  break;
 	}
 
       /* Print the block! */
       fprintf (fp, _("\nblock = %d\nsize  = %d\ncount = %d\n"), temp,
 	       av_stk->size, av_stk->count);
-      for (temp = 0; temp < av_stk->count; temp++)
-	{
-	  fprintf (fp, "  %15d   %10lu \n", av_stk->av_table[temp].av_size,
-		   (unsigned long) av_stk->av_table[temp].av_adr);
-	}
+      if (gdbm_avail_block_validate (dbf, av_stk) == 0)
+	av_table_display (av_stk->av_table, av_stk->count, fp);
+      else
+	terror (_("invalid avail_block"));
       temp = av_stk->next_block;
     }
   free (av_stk);
@@ -299,12 +318,12 @@ _gdbm_print_bucket_cache (FILE *fp, GDBM_FILE dbf)
   if (dbf->bucket_cache != NULL)
     {
       fprintf (fp,
-	_("Bucket Cache (size %zu):\n  Index:  Address  Changed  Data_Hash \n"),
+	_("Bucket Cache (size %zu):\n  Index:         Address  Changed  Data_Hash \n"),
 	 dbf->cache_size);
       for (index = 0; index < dbf->cache_size; index++)
 	{
 	  changed = dbf->bucket_cache[index].ca_changed;
-	  fprintf (fp, "  %5d:  %7lu %7s  %x\n",
+	  fprintf (fp, "  %5d:  %15lu %7s  %x\n",
 		   index,
 		   (unsigned long) dbf->bucket_cache[index].ca_adr,
 		   (changed ? _("True") : _("False")),
@@ -358,14 +377,7 @@ void
 open_handler (struct handler_param *param)
 {
   char *name = tildexpand (PARAM_STRING (param, 0));
-  if (file_name)
-    { 
-      gdbm_close (gdbm_file);
-      gdbm_file = NULL;
-      free (file_name);
-      file_name = NULL;
-    }
-  
+  closedb ();
   if (opendb (name) == 0)
     file_name = name;
   else
@@ -379,12 +391,7 @@ close_handler (struct handler_param *param)
   if (!gdbm_file)
     terror (_("nothing to close"));
   else
-    {
-      gdbm_close (gdbm_file);
-      gdbm_file = NULL;
-      free (file_name);
-      file_name = NULL;
-    }
+    closedb ();
 }
 
 
@@ -468,7 +475,7 @@ store_handler (struct handler_param *param)
   if (gdbm_store (gdbm_file,
 		  PARAM_DATUM (param, 0), PARAM_DATUM (param, 1),
 		  GDBM_REPLACE) != 0)
-    terror ("%s", _("Item not inserted."));
+    terror (_("Item not inserted: %s."), gdbm_db_strerror (gdbm_file));
 }
 
 /* first - begin iteration */
@@ -538,7 +545,7 @@ reorganize_handler (struct handler_param *param GDBM_ARG_UNUSED)
   if (gdbm_reorganize (gdbm_file))
     terror ("%s", _("Reorganization failed."));
   else
-    fprintf (param->fp, _("Reorganization succeeded."));
+    fprintf (param->fp, "%s\n", _("Reorganization succeeded."));
 }
 
 static void
@@ -552,7 +559,7 @@ err_printer (void *data GDBM_ARG_UNUSED, char const *fmt, ...)
   fprintf (stderr, "\n");
 }
 
-/* recover verbose backup max-failed-keys=N max-failed-buckets=N max-failures=N */
+/* recover sumamry verbose backup max-failed-keys=N max-failed-buckets=N max-failures=N */
 void
 recover_handler (struct handler_param *param)
 {
@@ -561,8 +568,9 @@ recover_handler (struct handler_param *param)
   int rc;
   int i;
   char *p;
+  int summary = 0;
   
-  for (i = 1; i < param->argc; i++)
+  for (i = 0; i < param->argc; i++)
     {
       char *arg = PARAM_STRING (param, i);
       if (strcmp (arg, "verbose") == 0)
@@ -570,6 +578,10 @@ recover_handler (struct handler_param *param)
 	  rcvr.errfun = err_printer;
 	  flags |= GDBM_RCVR_ERRFUN;
 	}
+      else if (strcmp (arg, "summary") == 0)
+	{
+	  summary = 1;
+	}
       else if (strcmp (arg, "backup") == 0)
 	{
 	  rcvr.errfun = err_printer;
@@ -617,6 +629,19 @@ recover_handler (struct handler_param *param)
   if (rc == 0)
     {
       fprintf (param->fp, _("Recovery succeeded.\n"));
+      if (summary)
+	{
+	  fprintf (param->fp,
+		   _("Keys recovered: %lu, failed: %lu, duplicate: %lu\n"),
+		   (unsigned long) rcvr.recovered_keys,
+		   (unsigned long) rcvr.failed_keys,
+		   (unsigned long) rcvr.duplicate_keys);
+	  fprintf (param->fp,
+		   _("Buckets recovered: %lu, failed: %lu\n"),
+		   (unsigned long) rcvr.recovered_buckets,
+		   (unsigned long) rcvr.failed_buckets);
+	}
+      
       if (rcvr.backup_name)
 	{
 	  fprintf (param->fp,
@@ -728,7 +753,11 @@ print_bucket_begin (struct handler_param *param, size_t *exp_count)
       terror (_("Not a bucket."));
       return 1;
     }
-  _gdbm_get_bucket (gdbm_file, temp);
+  if (_gdbm_get_bucket (gdbm_file, temp))
+    {
+      terror ("%s", gdbm_db_strerror (gdbm_file));
+      return 1;
+    }
   if (exp_count)
     *exp_count = bucket_print_lines (gdbm_file->bucket) + 3;
   return 0;
@@ -910,8 +939,7 @@ list_handler (struct handler_param *param)
 void
 quit_handler (struct handler_param *param GDBM_ARG_UNUSED)
 {
-  if (gdbm_file != NULL)
-    gdbm_close (gdbm_file);
+  closedb ();
   input_done ();
   exit (EXIT_OK);
 }
@@ -925,7 +953,7 @@ export_handler (struct handler_param *param)
   int i;
   int filemode;
 
-  for (i = 1; i < param->argc; i++)
+  for (i = 0; i < param->argc; i++)
     {
       if (strcmp (PARAM_STRING (param, i), "truncate") == 0)
 	 flags = GDBM_NEWDB;
@@ -959,7 +987,7 @@ import_handler (struct handler_param *param)
   int i;
   int rc;
 
-  for (i = 1; i < param->argc; i++)
+  for (i = 0; i < param->argc; i++)
     {
       if (strcmp (PARAM_STRING (param, i), "replace") == 0)
 	 flag = GDBM_REPLACE;
@@ -1106,9 +1134,10 @@ void
 source_handler (struct handler_param *param)
 {
   char *fname = tildexpand (PARAM_STRING (param, 0));
-  if (setsource (fname, 0) == 0)
-    yyparse ();
+  instream_t istr = instream_file_create (fname);
   free (fname);
+  if (istr && input_context_push (istr) == 0)
+    yyparse ();
 }
 
 
@@ -1148,19 +1177,19 @@ struct command
 struct command command_tab[] = {
 #define S(s) #s, sizeof (#s) - 1
   { S(count), T_CMD,
-    checkdb, count_handler, NULL,
+    checkdb_begin, count_handler, NULL,
     { { NULL } },
     FALSE,
     REPEAT_NEVER,
     N_("count (number of entries)") },
   { S(delete), T_CMD,
-    checkdb, delete_handler, NULL,
+    checkdb_begin, delete_handler, NULL,
     { { N_("KEY"), GDBM_ARG_DATUM, DS_KEY }, { NULL } },
     FALSE,
     REPEAT_NEVER,
     N_("delete a record") },
   { S(export), T_CMD,
-    checkdb, export_handler, NULL,
+    checkdb_begin, export_handler, NULL,
     { { N_("FILE"), GDBM_ARG_STRING },
       { "[truncate]", GDBM_ARG_STRING },
       { "[binary|ascii]", GDBM_ARG_STRING },
@@ -1169,7 +1198,7 @@ struct command command_tab[] = {
     REPEAT_NEVER,
     N_("export") },
   { S(fetch), T_CMD,
-    checkdb, fetch_handler, NULL,
+    checkdb_begin, fetch_handler, NULL,
     { { N_("KEY"), GDBM_ARG_DATUM, DS_KEY }, { NULL } },
     FALSE,
     REPEAT_NEVER,
@@ -1190,14 +1219,14 @@ struct command command_tab[] = {
     REPEAT_NEVER,
     N_("list") },
   { S(next), T_CMD,
-    checkdb, nextkey_handler, NULL,
+    checkdb_begin, nextkey_handler, NULL,
     { { N_("[KEY]"), GDBM_ARG_DATUM, DS_KEY },
       { NULL } },
     FALSE,
     REPEAT_NOARG,
     N_("nextkey") },
   { S(store), T_CMD,
-    checkdb, store_handler, NULL,
+    checkdb_begin, store_handler, NULL,
     { { N_("KEY"), GDBM_ARG_DATUM, DS_KEY },
       { N_("DATA"), GDBM_ARG_DATUM, DS_CONTENT },
       { NULL } },
@@ -1205,20 +1234,21 @@ struct command command_tab[] = {
     REPEAT_NEVER,
     N_("store") },
   { S(first), T_CMD,
-    checkdb, firstkey_handler, NULL,
+    checkdb_begin, firstkey_handler, NULL,
     { { NULL } },
     FALSE,
     REPEAT_NEVER,
     N_("firstkey") },
   { S(reorganize), T_CMD,
-    checkdb, reorganize_handler, NULL,
+    checkdb_begin, reorganize_handler, NULL,
     { { NULL } },
     FALSE,
     REPEAT_NEVER,
     N_("reorganize") },
   { S(recover), T_CMD,
-    checkdb, recover_handler, NULL,
+    checkdb_begin, recover_handler, NULL,
     { { "[verbose]", GDBM_ARG_STRING },
+      { "[summary]", GDBM_ARG_STRING },
       { "[backup]",  GDBM_ARG_STRING },
       { "[max-failed-keys=N]", GDBM_ARG_STRING },
       { "[max-failed-buckets=N]", GDBM_ARG_STRING },
@@ -1456,7 +1486,7 @@ command_lookup (const char *str, struct locus *loc, struct command **pcmd)
 	      break;
 
 	    case fcom_found:
-	      if (!interactive)
+	      if (!interactive ())
 		{
 		  state = fcom_abort;
 		  found = NULL;
@@ -1479,9 +1509,8 @@ command_lookup (const char *str, struct locus *loc, struct command **pcmd)
     }
 
   if (state == fcom_init)
-    lerror (loc,
-		 interactive ? _("Invalid command. Try ? for help.") :
-	                       _("Unknown command"));
+    lerror (loc, interactive () ? _("Invalid command. Try ? for help.") :
+	                          _("Unknown command"));
   if (!found)
     return T_BOGUS;
 
@@ -1490,7 +1519,12 @@ command_lookup (const char *str, struct locus *loc, struct command **pcmd)
 }
 
 char *parseopt_program_doc = N_("examine and/or modify a GDBM database");
-char *parseopt_program_args = N_("DBFILE");
+char *parseopt_program_args = N_("DBFILE [COMMAND [ARG ...]]");
+
+enum {
+  OPT_LEX_TRACE = 256,
+  OPT_GRAM_TRACE
+};
 
 struct gdbm_option optab[] = {
   { 'b', "block-size", N_("SIZE"), N_("set block size") },
@@ -1504,6 +1538,10 @@ struct gdbm_option optab[] = {
   { 'r', "read-only",  NULL,       N_("open database in read-only mode") },
   { 's', "synchronize", NULL,      N_("synchronize to disk after each write") },
   { 'q', "quiet",      NULL,       N_("don't print initial banner") },
+#if GDBMTOOL_DEBUG    
+  { OPT_LEX_TRACE, "lex-trace", NULL, N_("enable lexical analyzer traces") },
+  { OPT_GRAM_TRACE, "gram-trace", NULL, N_("enable grammar traces") },
+#endif  
   { 0 }
 };
 
@@ -1786,7 +1824,7 @@ static struct gdbmarglist last_args;
 void
 run_last_command (void)
 {
-  if (interactive)
+  if (interactive ())
     {
       if (last_cmd)
 	{
@@ -1846,7 +1884,7 @@ run_command (struct command *cmd, struct gdbmarglist *arglist)
 	/* Optional argument */
 	break;
 
-      if (!interactive)
+      if (!interactive ())
 	{
 	  terror (_("%s: not enough arguments"), cmd->name);
 	  return 1;
@@ -1896,7 +1934,7 @@ run_command (struct command *cmd, struct gdbmarglist *arglist)
   pagfp = NULL;
       
   expected_lines = 0;
-  expected_lines_ptr = (interactive && pager) ? &expected_lines : NULL;
+  expected_lines_ptr = (interactive () && pager) ? &expected_lines : NULL;
   if (!(cmd->begin && cmd->begin (&param, expected_lines_ptr)))
     {
       if (pager && expected_lines > get_screen_lines ())
@@ -1938,12 +1976,13 @@ run_command (struct command *cmd, struct gdbmarglist *arglist)
 }
 
 static void
-source_rcfile ()
+source_rcfile (void)
 {
+  instream_t istr = NULL;
+  
   if (access (GDBMTOOLRC, R_OK) == 0)
     {
-      if (setsource (GDBMTOOLRC, 0) == 0)
-	yyparse ();
+      istr = instream_file_create (GDBMTOOLRC);
     }
   else
     {
@@ -1962,11 +2001,17 @@ source_rcfile ()
       fname = mkfilename (p, GDBMTOOLRC, NULL);
       if (access (fname, R_OK) == 0)
 	{
-	  if (setsource (fname, 0) == 0)
-	    yyparse ();
+	  istr = instream_file_create (GDBMTOOLRC);
 	}
       free (fname);
     }
+
+  if (istr)
+    {
+      if (input_context_push (istr))
+	exit (EXIT_FATAL);
+      yyparse ();
+    }
 }
 
 #if GDBM_DEBUG_ENABLE
@@ -1984,12 +2029,12 @@ debug_printer (char const *fmt, ...)
 int
 main (int argc, char *argv[])
 {
-  int intr;  
   int opt;
   int bv;
   int norc = 0;
   int res;
-  char *source = "-";
+  char *source = NULL;
+  instream_t input = NULL;
   
   set_progname (argv[0]);
 #if GDBM_DEBUG_ENABLE
@@ -2005,8 +2050,6 @@ main (int argc, char *argv[])
   sort_commands ();
   
   /* Initialize variables. */
-  intr = isatty (0);
-  interactive = intr; /* Used early by input_init */
   dsdef[DS_KEY] = dsegm_new_field (datadef_lookup ("string"), NULL, 1);
   dsdef[DS_CONTENT] = dsegm_new_field (datadef_lookup ("string"), NULL, 1);
 
@@ -2014,6 +2057,7 @@ main (int argc, char *argv[])
   variable_set ("pager", VART_STRING, getenv ("PAGER"));
   
   input_init ();
+  lex_trace (0);
   
   for (opt = parseopt_first (argc, argv, optab);
        opt != EOF;
@@ -2022,7 +2066,6 @@ main (int argc, char *argv[])
       {
       case 'f':
 	source = optarg;
-	intr = 0;
 	break;
 	
       case 'l':
@@ -2068,6 +2111,14 @@ main (int argc, char *argv[])
 	bv = 1;
 	variable_set ("quiet", VART_BOOL, &bv);
 	break;
+
+      case OPT_LEX_TRACE:
+	lex_trace (1);
+	break;
+
+      case OPT_GRAM_TRACE:
+	gram_trace (1);
+	break;
 	
       default:
 	terror (_("unknown option; try `%s -h' for more info"),
@@ -2078,14 +2129,30 @@ main (int argc, char *argv[])
   argc -= optind;
   argv += optind;
 
-  if (argc > 1)
+  if (source && strcmp (source, "-"))
     {
-      terror (_("too many arguments"));
-      exit (EXIT_USAGE);
+      input = instream_file_create (source);
+      if (!input)
+	exit (1);
+    }
+  
+  if (argc >= 1)
+    {
+      file_name = estrdup (argv[0]);
+      argc--;
+      argv++;
+      if (argc)
+	{
+	  if (input)
+	    {
+	      terror (_("--file and and command cannot be used together"));
+	      exit (EXIT_USAGE);
+	    }
+	  input = instream_argv_create (argc, argv);
+	  if (!input)
+	    exit (1);
+	}
     }
-      
-  if (argc == 1)
-    file_name = estrdup (argv[0]);
 
   signal (SIGPIPE, SIG_IGN);
 
@@ -2093,15 +2160,19 @@ main (int argc, char *argv[])
   argmax = 0;
 
   if (!norc)
-    source_rcfile ();	  
+    source_rcfile ();
+
+  if (!input)
+    input = instream_stdin_create ();
   
   /* Welcome message. */
-  if (intr && !variable_is_true ("quiet"))
+  if (instream_interactive (input) && !variable_is_true ("quiet"))
     printf (_("\nWelcome to the gdbm tool.  Type ? for help.\n\n"));
 
-  if (setsource (source, intr))
+  if (input_context_push (input))
     exit (EXIT_FATAL);
   res = yyparse ();
+  closedb ();
   input_done ();
   return res;
 }
